=== modified file 'NEWS'
--- NEWS	2008-11-20 04:17:17 +0000
+++ NEWS	2008-11-20 04:45:01 +0000
@@ -31,7 +31,9 @@
       (Martin Pool, #289148)
 
     * Don't call the system ``chdir()`` with an empty path. Sun OS seems
-      to give an error in that case. (John Arbash Meinel, #297831)
+      to give an error in that case.  Also, don't count on ``getcwd()`` 
+      being able to allocate a new buffer, which is a gnu extension.  
+      (John Arbash Meinel, Martin Pool, Harry Hirsch, #297831)
 
     * TooManyConcurrentRequests no longer occur when a fetch fails and
       tries to abort a write group.  This allows the root cause (e.g. a

=== modified file 'bzrlib/_readdir_pyx.pyx'
--- bzrlib/_readdir_pyx.pyx	2008-11-20 04:17:17 +0000
+++ bzrlib/_readdir_pyx.pyx	2008-11-20 04:45:01 +0000
@@ -36,6 +36,8 @@
 
 cdef extern from 'unistd.h':
     int chdir(char *path)
+    int close(int fd)
+    int fchdir(int fd)
     char *getcwd(char *, int size)
 
 cdef extern from 'stdlib.h':
@@ -49,6 +51,7 @@
     ctypedef long time_t
     ctypedef unsigned long ino_t
     ctypedef unsigned long long off_t
+    ctypedef int mode_t
 
 
 cdef extern from 'sys/stat.h':
@@ -69,6 +72,11 @@
     int S_ISSOCK(int mode)
 
 
+cdef extern from 'fcntl.h':
+    int O_RDONLY
+    int open(char *pathname, int flags, mode_t mode)
+
+
 cdef extern from 'Python.h':
     char * PyString_AS_STRING(object)
     ctypedef int Py_ssize_t # Required for older pyrex versions
@@ -279,14 +287,23 @@
     cdef char *name
     cdef int stat_result
     cdef _Stat statvalue
-    cdef char *cwd
     global errno
+    cdef int orig_dir_fd
 
-    cwd = getcwd(NULL, 0)
-    if path != "":
-        # Avoid chdir('') because it causes problems on Sun OS
+    # Avoid chdir('') because it causes problems on Sun OS, and avoid this if
+    # staying in .
+    if path != "" and path != '.':
+        # we change into the requested directory before reading, and back at the
+        # end, because that turns out to make the stat calls measurably faster than
+        # passing full paths every time.
+        orig_dir_fd = open(".", O_RDONLY, 0)
+        if orig_dir_fd == -1:
+            raise OSError(errno, strerror(errno))
         if -1 == chdir(path):
             raise OSError(errno, strerror(errno))
+    else:
+        orig_dir_fd = -1
+
     try:
         the_dir = opendir(".")
         if NULL == the_dir:
@@ -340,10 +357,12 @@
             if -1 == closedir(the_dir):
                 raise OSError(errno, strerror(errno))
     finally:
-        if -1 == chdir(cwd):
-            free(cwd)
-            raise OSError(errno, strerror(errno))
-        free(cwd)
+        if -1 != orig_dir_fd:
+            if -1 == fchdir(orig_dir_fd):
+                raise OSError(errno, strerror(errno))
+            if -1 == close(orig_dir_fd):
+                raise OSError(errno, strerror(errno))
+
     return result
 
 

